MALE_NAME = "JEFF"
MALE_NAME2 = "BIFF"
FEMALE_NAME = "JENNIFER"
FEMALE_NAME2 = "JANE"
MALE_SURNAME = "BRIDGES"
FEMALE_SURNAME = "SAUNDERS"
MALE_VALUE = "M"
FEMALE_VALUE = "F"
ADDRESS_LINE_1 = "15 THE DRIVE"
ADDRESS_LINE_2 = "SOMEWHERETON"
ADDRESS_LINE_3 = "SOMEWHERESHIRE"
ADDRESS_POSTCODE = "SW1 3AB"
PHONE_NUMBER = "01234 567890"